""" The grammar of low-level values in the GD file format """

from pyparsing import (
    Forward,
    Group,
    Keyword,
    Optional,
    QuotedString,
    Suppress,
    Word,
    alphanums,
    alphas,
    delimitedList,
    pyparsing_common,
)

from .objects import GDObject

boolean = (
    (Keyword("true") | Keyword("false"))
    .setName("bool")
    .setParseAction(lambda x: x[0].lower() == "true")
)

null = Keyword("null").setParseAction(lambda _: [None])


primitive = (
    null
    | QuotedString('"', escChar="\\", multiline=True)
    | boolean
    | pyparsing_common.number
)
value = Forward()

# Vector2( 1, 2 )
obj_type = (
    Word(alphas, alphanums).setResultsName("object_name")
    + Suppress("(")
    + delimitedList(value)
    + Suppress(")")
).setParseAction(GDObject.from_parser)

# [ 1, 2 ]
list_ = (
    Group(Suppress("[") + Optional(delimitedList(value)) + Suppress("]"))
    .setName("list")
    .setParseAction(lambda p: p.asList())
)
key_val = Group(QuotedString('"', escChar="\\") + Suppress(":") + value)

# {
# "_edit_use_anchors_": false
# }
dict_ = (
    (Suppress("{") + Optional(delimitedList(key_val)) + Suppress("}"))
    .setName("dict")
    .setParseAction(lambda d: {k: v for k, v in d})
)

# Exports

value <<= primitive | list_ | dict_ | obj_type



import re
from collections import OrderedDict
from typing import Optional, Type, TypeVar

from .objects import ExtResource
from .util import stringify_object

__all__ = [
    "GDSectionHeader",
    "GDSection",
    "GDNodeSection",
    "GDExtResourceSection",
    "GDSubResourceSection",
    "GDResourceSection",
]

GDSectionHeaderType = TypeVar("GDSectionHeaderType", bound="GDSectionHeader")

GD_SECTION_REGISTRY = {}


class GDSectionHeader(object):
    """
    Represents the header for a section

    example::

        [node name="Sprite" type="Sprite" index="3"]
    """

    def __init__(self, _name: str, **kwargs) -> None:
        self.name = _name
        self.attributes = OrderedDict()
        for k, v in kwargs.items():
            self.attributes[k] = v

    def __getitem__(self, k: str):
        return self.attributes[k]

    def __setitem__(self, k: str, v):
        self.attributes[k] = v

    def __delitem__(self, k: str):
        try:
            del self.attributes[k]
        except KeyError:
            pass

    def get(self, k: str, default=None):
        return self.attributes.get(k, default)

    @classmethod
    def from_parser(
        cls: Type[GDSectionHeaderType], parse_result
    ) -> GDSectionHeaderType:
        header = cls(parse_result[0])
        for attribute in parse_result[1:]:
            header.attributes[attribute[0]] = attribute[1]
        return header

    def __str__(self) -> str:
        attribute_str = ""
        if self.attributes:
            attribute_str = " " + " ".join(
                ["%s=%s" % (k, stringify_object(v)) for k, v in self.attributes.items()]
            )
        return "[" + self.name + attribute_str + "]"

    def __repr__(self) -> str:
        return "GDSectionHeader(%s)" % self.__str__()

    def __eq__(self, other) -> bool:
        if not isinstance(other, GDSectionHeader):
            return False
        return self.name == other.name and self.attributes == other.attributes

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)


class GDSectionMeta(type):
    """ Still trying to be too clever """

    def __new__(cls, name, bases, dct):
        x = super().__new__(cls, name, bases, dct)
        section_name_camel = name[2:-7]
        section_name = re.sub(r"(?<!^)(?=[A-Z])", "_", section_name_camel).lower()
        GD_SECTION_REGISTRY[section_name] = x
        return x


GDSectionType = TypeVar("GDSectionType", bound="GDSection")


class GDSection(metaclass=GDSectionMeta):
    """
    Represents a full section of a GD file

    example::

        [node name="Sprite" type="Sprite"]
        texture = ExtResource( 1 )

    """

    def __init__(self, header: GDSectionHeader, **kwargs) -> None:
        self.header = header
        self.properties = OrderedDict()
        for k, v in kwargs.items():
            self.properties[k] = v

    def __getitem__(self, k: str):
        return self.properties[k]

    def __setitem__(self, k: str, v):
        self.properties[k] = v

    def __delitem__(self, k: str):
        try:
            del self.properties[k]
        except KeyError:
            pass

    def get(self, k: str, default=None):
        return self.properties.get(k, default)

    @classmethod
    def from_parser(cls: Type[GDSectionType], parse_result):
        header = parse_result[0]
        factory = GD_SECTION_REGISTRY.get(header.name, cls)
        section = factory.__new__(factory)
        section.header = header
        section.properties = OrderedDict()
        for k, v in parse_result[1:]:
            section[k] = v
        return section

    def __str__(self) -> str:
        ret = str(self.header)
        if self.properties:
            ret += "\n" + "\n".join(
                [
                    "%s = %s" % (k, stringify_object(v))
                    for k, v in self.properties.items()
                ]
            )
        return ret

    def __repr__(self) -> str:
        return "%s(%s)" % (type(self).__name__, self.__str__())

    def __eq__(self, other) -> bool:
        if not isinstance(other, GDSection):
            return False
        return (
            self.header.name == other.header.name
            and self.properties == other.properties
        )

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)


class GDExtResourceSection(GDSection):
    """ Section representing an [ext_resource] """

    def __init__(self, path: str, type: str, id: int):
        super().__init__(GDSectionHeader("ext_resource", path=path, type=type, id=id))

    @property
    def path(self) -> str:
        return self.header["path"]

    @path.setter
    def path(self, path: str) -> None:
        self.header["path"] = path

    @property
    def type(self) -> str:
        return self.header["type"]

    @type.setter
    def type(self, type: str) -> None:
        self.header["type"] = type

    @property
    def id(self) -> int:
        return self.header["id"]

    @id.setter
    def id(self, id: int) -> None:
        self.header["id"] = id


class GDSubResourceSection(GDSection):
    """ Section representing a [sub_resource] """

    def __init__(self, type: str, id: int, **kwargs):
        super().__init__(GDSectionHeader("sub_resource", type=type, id=id), **kwargs)

    @property
    def type(self) -> str:
        return self.header["type"]

    @type.setter
    def type(self, type: str) -> None:
        self.header["type"] = type

    @property
    def id(self) -> int:
        return self.header["id"]

    @id.setter
    def id(self, id: int) -> None:
        self.header["id"] = id


class GDNodeSection(GDSection):
    """ Section representing a [node] """

    def __init__(
        self,
        name: str,
        type: str = None,
        parent: str = None,
        instance: int = None,
        index: int = None,
        # TODO: instance_placeholder, owner, groups are referenced in the docs, but I
        # haven't seen them come up yet in my project
    ):
        kwargs = {
            "name": name,
            "type": type,
            "parent": parent,
            "instance": ExtResource(instance) if instance is not None else None,
            "index": str(index) if index is not None else None,
        }
        super().__init__(
            GDSectionHeader(
                "node", **{k: v for k, v in kwargs.items() if v is not None}
            )
        )

    @classmethod
    def ext_node(cls, name: str, instance: int, parent: str = None, index: int = None):
        return cls(name, parent=parent, instance=instance, index=index)

    @property
    def name(self) -> str:
        return self.header["name"]

    @name.setter
    def name(self, name: str) -> None:
        self.header["name"] = name

    @property
    def type(self) -> Optional[str]:
        return self.header.get("type")

    @type.setter
    def type(self, type: Optional[str]) -> None:
        if type is None:
            del self.header["type"]
        else:
            self.header["type"] = type
            self.instance = None

    @property
    def parent(self) -> Optional[str]:
        return self.header.get("parent")

    @parent.setter
    def parent(self, parent: Optional[str]) -> None:
        if parent is None:
            del self.header["parent"]
        else:
            self.header["parent"] = parent

    @property
    def instance(self) -> Optional[int]:
        resource = self.header.get("instance")
        if resource is not None:
            return resource.id
        return None

    @instance.setter
    def instance(self, instance: Optional[int]) -> None:
        if instance is None:
            del self.header["instance"]
        else:
            self.header["instance"] = ExtResource(instance)
            self.type = None

    @property
    def index(self) -> Optional[int]:
        idx = self.header.get("index")
        if idx is not None:
            return int(idx)
        return None

    @index.setter
    def index(self, index: Optional[int]) -> None:
        if index is None:
            del self.header["index"]
        else:
            self.header["index"] = str(index)


class GDResourceSection(GDSection):
    """ Represents a [resource] section """

    def __init__(self, **kwargs):
        super().__init__(GDSectionHeader("resource"), **kwargs)



""" Wrappers for Godot's non-primitive object types """

from functools import partial
from typing import Type, TypeVar

from .util import stringify_object

__all__ = []

GD_OBJECT_REGISTRY = {}


class GDObjectMeta(type):
    """
    This is me trying to be too clever for my own good

    Odds are high that it'll cause some weird hard-to-debug issues at some point, but
    isn't it neeeeeat? -_-
    """

    def __new__(cls, name, bases, dct):
        x = super().__new__(cls, name, bases, dct)
        GD_OBJECT_REGISTRY[name] = x
        __all__.append(name)
        return x


GDObjectType = TypeVar("GDObjectType", bound="GDObject")


class GDObject(metaclass=GDObjectMeta):
    """
    Base class for all GD Object types

    Can be used to represent any GD type. For example::

        GDObject('Vector2', 1, 2) == Vector2(1, 2)
    """

    def __init__(self, name, *args) -> None:
        self.name = name
        self.args = list(args)

    @classmethod
    def from_parser(cls: Type[GDObjectType], parse_result) -> GDObjectType:
        name = parse_result[0]
        factory = GD_OBJECT_REGISTRY.get(name, partial(GDObject, name))
        return factory(*parse_result[1:])

    def __str__(self) -> str:
        return "%s( %s )" % (
            self.name,
            ", ".join([stringify_object(v) for v in self.args]),
        )

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other) -> bool:
        if not isinstance(other, GDObject):
            return False
        return self.name == other.name and self.args == other.args

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)


class Vector2(GDObject):
    def __init__(self, x: float, y: float) -> None:
        super().__init__("Vector2", x, y)

    def __getitem__(self, idx) -> float:
        return self.args[idx]

    def __setitem__(self, idx: int, value: float):
        self.args[idx] = value

    @property
    def x(self) -> float:
        """ Getter for x """
        return self.args[0]

    @x.setter
    def x(self, x: float) -> None:
        """ Setter for x """
        self.args[0] = x

    @property
    def y(self) -> float:
        """ Getter for y """
        return self.args[1]

    @y.setter
    def y(self, y: float) -> None:
        """ Setter for y """
        self.args[1] = y


class Vector3(GDObject):
    def __init__(self, x: float, y: float, z: float) -> None:
        super().__init__("Vector3", x, y, z)

    def __getitem__(self, idx: int) -> float:
        return self.args[idx]

    def __setitem__(self, idx: int, value: float) -> None:
        self.args[idx] = value

    @property
    def x(self) -> float:
        """ Getter for x """
        return self.args[0]

    @x.setter
    def x(self, x: float) -> None:
        """ Setter for x """
        self.args[0] = x

    @property
    def y(self) -> float:
        """ Getter for y """
        return self.args[1]

    @y.setter
    def y(self, y: float) -> None:
        """ Setter for y """
        self.args[1] = y

    @property
    def z(self) -> float:
        """ Getter for z """
        return self.args[2]

    @z.setter
    def z(self, z: float) -> None:
        """ Setter for z """
        self.args[2] = z


class Color(GDObject):
    def __init__(self, r: float, g: float, b: float, a: float) -> None:
        assert 0 <= r <= 1
        assert 0 <= g <= 1
        assert 0 <= b <= 1
        assert 0 <= a <= 1
        super().__init__("Color", r, g, b, a)

    def __getitem__(self, idx: int) -> float:
        return self.args[idx]

    def __setitem__(self, idx: int, value: float) -> None:
        self.args[idx] = value

    @property
    def r(self) -> float:
        """ Getter for r """
        return self.args[0]

    @r.setter
    def r(self, r: float) -> None:
        """ Setter for r """
        self.args[0] = r

    @property
    def g(self) -> float:
        """ Getter for g """
        return self.args[1]

    @g.setter
    def g(self, g: float) -> None:
        """ Setter for g """
        self.args[1] = g

    @property
    def b(self) -> float:
        """ Getter for b """
        return self.args[2]

    @b.setter
    def b(self, b: float) -> None:
        """ Setter for b """
        self.args[2] = b

    @property
    def a(self) -> float:
        """ Getter for a """
        return self.args[3]

    @a.setter
    def a(self, a: float) -> None:
        """ Setter for a """
        self.args[3] = a


class NodePath(GDObject):
    def __init__(self, path: str) -> None:
        super().__init__("NodePath", path)

    @property
    def path(self) -> str:
        """ Getter for path """
        return self.args[0]

    @path.setter
    def path(self, path: str) -> None:
        """ Setter for path """
        self.args[0] = path

    def __str__(self) -> str:
        return '%s("%s")' % (self.name, self.path)


class ExtResource(GDObject):
    def __init__(self, id: int) -> None:
        super().__init__("ExtResource", id)

    @property
    def id(self) -> int:
        """ Getter for id """
        return self.args[0]

    @id.setter
    def id(self, id: int) -> None:
        """ Setter for id """
        self.args[0] = id


class SubResource(GDObject):
    def __init__(self, id: int) -> None:
        super().__init__("SubResource", id)

    @property
    def id(self) -> int:
        """ Getter for id """
        return self.args[0]

    @id.setter
    def id(self, id: int) -> None:
        """ Setter for id """
        self.args[0] = id



from .files import *
from .objects import *
from .sections import *
from .tree import *

__version__ = "0.1"


def parse(string: str) -> GDFile:
    return GDFile.parse(string)


def load(fp) -> GDFile:
    return parse(fp.read())



""" Utils """


def stringify_object(value):
    """ Serialize a value to the godot file format """
    if value is None:
        return "null"
    elif isinstance(value, str):
        return '"' + value + '"'
    elif isinstance(value, bool):
        return "true" if value else "false"
    elif isinstance(value, dict):
        return (
            "{\n"
            + ",\n".join(
                ['"%s": %s' % (k, stringify_object(v)) for k, v in value.items()]
            )
            + "\n}"
        )
    elif isinstance(value, list):
        return "[ " + ", ".join([stringify_object(v) for v in value]) + " ]"
    else:
        return str(value)



""" Helper API for working with the Godot scene tree structure """
from collections import OrderedDict
from typing import Union

from .sections import GDNodeSection

__all__ = ["Node"]


class Node(object):
    """
    Wraps a GDNodeSection object

    Provides a way to access the node sections that is spatially-aware. It is stored in
    a tree structure instead of the flat list that the file format demands.
    """

    def __init__(
        self,
        name: str,
        type: str = None,
        instance: int = None,
        section: GDNodeSection = None,
        properties: dict = None,
    ):
        self.name = name
        self.type = type
        self.instance = instance
        self.parent = None
        self.section = section or GDNodeSection(name)
        self.properties = (
            OrderedDict() if properties is None else OrderedDict(properties)
        )
        self._children = []  # type: ignore

    def __getitem__(self, k: str):
        return self.properties[k]

    def __setitem__(self, k: str, v):
        self.properties[k] = v

    def __delitem__(self, k: str):
        try:
            del self.properties[k]
        except KeyError:
            pass

    def get(self, k: str, default=None):
        return self.properties.get(k, default)

    @classmethod
    def from_section(cls, section: GDNodeSection):
        """ Create a Node from a GDNodeSection """
        return cls(
            section.name,
            section.type,
            section.instance,
            section,
            properties=section.properties,
        )

    def flatten(self, path: str = None):
        """
        Write values to GDNodeSection and iterate over children

        This call will copy the existing values on this node into the GDNodeSection and
        iterate over self and all child nodes, calling flatten on them as well.
        """

        self._update_section(path)

        yield self
        if path is None:
            child_path = "."
        elif path == ".":
            child_path = self.name
        else:
            child_path = path + "/" + self.name
        for child in self._children:
            yield from child.flatten(child_path)

    def _update_section(self, path: str = None):
        self.section.name = self.name
        self.section.type = self.type
        self.section.parent = path
        self.section.instance = self.instance
        self.section.properties = self.properties
        return self.section

    def get_children(self):
        """ Get all children of this node """
        return self._children

    def get_child(self, name_or_index: Union[int, str]):
        """ Get a child by name or index """
        if isinstance(name_or_index, int):
            return self._children[name_or_index]
        for node in self._children:
            if node.name == name_or_index:
                return node

    def get_node(self, path):
        """ Mimics the Godot get_node() behavior """
        if path in (".", ""):
            return self
        pieces = path.split("/")
        child = self.get_child(pieces[0])
        if child is None:
            return None
        return child.get_node("/".join(pieces[1:]))

    def add_child(self, node):
        """ Add a child to the current node """
        self._children.append(node)
        node.parent = self

    def remove_from_parent(self):
        """ Remove this node from its parent """
        self.parent.remove_child(self)

    def remove_child(self, node_or_name_or_index):
        """
        Remove a child

        You can pass in a Node, the name of a Node, or the index of the child
        """
        if isinstance(node_or_name_or_index, str):
            for i, node in enumerate(self._children):
                if node.name == node_or_name_or_index:
                    child = self._children.pop(i)
                    return
        elif isinstance(node_or_name_or_index, int):
            child = self._children.pop(node_or_name_or_index)
        else:
            self._children.remove(node_or_name_or_index)
            child = node_or_name_or_index
        if child is not None:
            child.parent = None

    def __str__(self):
        return "Node(%s)" % self.name

    def __repr__(self):
        return str(self)


class Tree(object):
    """ Container for the scene tree """

    def __init__(self, root=None):
        self.root = root

    @classmethod
    def build(cls, nodes):
        """ Build the Tree from a flat list of [node]'s """
        tree = cls()
        # Makes assumptions that the nodes are well-ordered
        for node in nodes:
            if node.parent is None:
                tree.root = Node.from_section(node)
            else:
                parent = tree.root.get_node(node.parent)
                parent.add_child(Node.from_section(node))
        return tree

    def flatten(self):
        """ Flatten the tree back into a list of GDNodeSection """
        ret = []
        if self.root is None:
            return ret
        for node in self.root.flatten():
            ret.append(node.section)
        return ret



""" The grammar of the larger structures in the GD file format """

from pyparsing import (
    Empty,
    Group,
    Optional,
    Suppress,
    Word,
    alphanums,
    delimitedList,
    lineStart,
)

from .sections import GDSection, GDSectionHeader
from .values import value

key = Word(alphanums + "_/").setName("key")
var = Word(alphanums + "_").setName("variable")
attribute = Group(var + Suppress("=") + value)

# [node name="Node2D"]
section_header = (
    (
        Suppress(lineStart)
        + Suppress("[")
        + var.setResultsName("section_type")
        + Optional(delimitedList(attribute, Empty()))
        + Suppress("]")
    )
    .setName("section_header")
    .setParseAction(GDSectionHeader.from_parser)
)

# texture = ExtResource( 1 )
section_entry = Group(Suppress(lineStart) + key + Suppress("=") + value).setName(
    "section_entry"
)
section_contents = delimitedList(section_entry, Empty()).setName("section_contents")

# [node name="Sprite" type="Sprite"]
# texture = ExtResource( 1 )
section = (
    (section_header + Optional(section_contents))
    .setName("section")
    .setParseAction(GDSection.from_parser)
)

# Exports

scene_file = delimitedList(section, Empty())



""" Setup file """
import os
import re

from setuptools import find_packages, setup

HERE = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(HERE, "README.rst")).read()
CHANGES = open(os.path.join(HERE, "CHANGES.rst")).read()
REQUIREMENTS_TEST = open(os.path.join(HERE, "requirements_test.txt")).readlines()

REQUIREMENTS = ["pyparsing"]


if __name__ == "__main__":
    setup(
        name="godot_parser",
        version="0.1",
        description="Python library for parsing Godot scene files",
        long_description=README + "\n\n" + CHANGES,
        long_description_content_type="text/x-rst",
        classifiers=[
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
        ],
        license="MIT",
        author="Steven Arcangeli",
        author_email="stevearc@stevearc.com",
        url="https://github.com/stevearc/godot_parser",
        keywords="godot parse parser scene",
        platforms="any",
        zip_safe=False,
        include_package_data=True,
        packages=find_packages(exclude=("tests",)),
        install_requires=REQUIREMENTS,
        tests_require=REQUIREMENTS + REQUIREMENTS_TEST,
        test_suite="tests",
    )



import os
from contextlib import contextmanager
from typing import List, Optional, Type, TypeVar

from .objects import ExtResource, GDObject, SubResource
from .sections import (
    GDExtResourceSection,
    GDNodeSection,
    GDSection,
    GDSectionHeader,
    GDSubResourceSection,
)
from .structure import scene_file
from .tree import Tree

__all__ = ["GDFile", "GDScene", "GDResource"]

# Scene and resource files seem to group the section types together and sort them.
# This is the order I've observed
SCENE_ORDER = [
    "gd_scene",
    "gd_resource",
    "ext_resource",
    "sub_resource",
    "resource",
    "node",
    "connection",
    "editable",
]


GDFileType = TypeVar("GDFileType", bound="GDFile")


class GDFile(object):
    """ Base class representing the contents of a Godot file """

    def __init__(self, *sections: GDSection) -> None:
        self._sections = list(sections)

    def add_section(self, new_section: GDSection) -> int:
        """ Add a section to the file and return the index of that section """
        new_idx = SCENE_ORDER.index(new_section.header.name)
        for i, section in enumerate(self._sections):
            idx = SCENE_ORDER.index(section.header.name)
            if new_idx < idx:  # type: ignore
                self._sections.insert(i, new_section)
                return i
        self._sections.append(new_section)
        return len(self._sections) - 1

    def remove_section(self, section: GDSection) -> bool:
        """ Remove a section from the file """
        idx = -1
        for i, s in enumerate(self._sections):
            if section == s:
                idx = i
                break
        if idx == -1:
            return False
        self.remove_at(idx)
        return True

    def remove_at(self, index: int):
        """ Remove a section at an index """
        return self._sections.pop(index)

    def get_sections(self, name: str = None) -> List[GDSection]:
        """ Get all sections, or all sections of a given type """
        if name is None:
            return self._sections
        return [s for s in self._sections if s.header.name == name]

    def find_section(self, section_name_: str = None, **constraints):
        """
        Find the first section that matches

        You may pass in a section_name, which will match the header name (e.g. 'node').
        You may also pass in kwargs that act as filters. For example::

            # Find the first node
            scene.find_section('node')
            # Find the first Sprite
            scene.find_section('node', type='Sprite')
            # Find the first ext_resource that references Health.tscn
            scene.find_section('ext_resourcee', path='Health.tscn')
        """
        for section in self.find_all(section_name_, **constraints):
            return section
        return None

    def find_all(self, section_name_: str = None, **constraints):
        """ Same as find_section, but returns all matches """
        for section in self.get_sections(section_name_):
            found = True
            for k, v in constraints.items():
                if getattr(section, k, None) == v:
                    continue
                if section.header.get(k) == v:
                    continue
                found = False
                break
            if found:
                yield section

    def add_ext_resource(self, path: str, type: str) -> int:
        """ Add an ext_resource """
        next_id = 1 + max(
            [s.id for s in self.get_sections("ext_resource")]  # type: ignore
            + [0]
        )
        self.add_section(GDExtResourceSection(path, type, next_id))
        return next_id

    def add_sub_resource(self, type: str, **kwargs) -> int:
        """ Add a sub_resource """
        next_id = 1 + max(
            [s.id for s in self.get_sections("sub_resource")]  # type: ignore
            + [0]
        )
        self.add_section(GDSubResourceSection(type, next_id, **kwargs))
        return next_id

    def add_node(
        self, name: str, type: str = None, parent: str = None, index: int = None,
    ) -> GDNodeSection:
        """
        Simple API for adding a node

        For a friendlier, tree-oriented API use use_tree()
        """
        node = GDNodeSection(name, type=type, parent=parent, index=index)
        self.add_section(node)
        return node

    def add_ext_node(
        self, name: str, instance: int, parent: str = None, index: int = None,
    ) -> GDNodeSection:
        """
        Simple API for adding a node that instances an ext_resource

        For a friendlier, tree-oriented API use use_tree()
        """
        node = GDNodeSection.ext_node(name, instance, parent=parent, index=index)
        self.add_section(node)
        return node

    @contextmanager
    def use_tree(self):
        """
        Helper API for working with the nodes in a tree structure

        This temporarily builds the nodes into a tree, and flattens them back into the
        GD file format when done.

        Example::

            with scene.use_tree() as tree:
                tree.root = Node('MyScene')
                tree.root.add_child(Node('Sensor', type='Area2D'))
                tree.root.add_child(Node('HealthBar', instance=1))
            scene.write("MyScene.tscn")
        """
        nodes = self.get_sections("node")
        tree = Tree.build(nodes)
        yield tree
        for i in range(len(self._sections) - 1, -1, -1):
            section = self._sections[i]
            if section.header.name == "node":
                self._sections.pop(i)
        nodes = tree.flatten()
        if not nodes:
            return
        # Let's find out where the root node belongs and then bulk add the rest at that
        # index
        i = self.add_section(nodes[0])
        self._sections[i + 1 : i + 1] = nodes[1:]

    def get_node(self, path: str = ".") -> Optional[GDNodeSection]:
        """ Mimics the Godot get_node API """
        with self.use_tree() as tree:
            if tree.root is None:
                return None
            node = tree.root.get_node(path)
        return node.section if node is not None else None

    @classmethod
    def parse(cls: Type[GDFileType], contents: str):
        """ Parse the contents of a Godot file """
        return cls.from_parser(scene_file.parseString(contents, parseAll=True))

    @classmethod
    def from_parser(cls: Type[GDFileType], parse_result):
        first_section = parse_result[0]
        if first_section.header.name == "gd_scene":
            scene = GDScene.__new__(GDScene)
            scene._sections = list(parse_result)
            return scene
        elif first_section.header.name == "gd_resource":
            resource = GDResource.__new__(GDResource)
            resource._sections = list(parse_result)
            return resource
        return cls(*parse_result)

    def write(self, filename):
        """ Writes this to a file """
        ext = os.path.splitext(filename)[1]
        with open(filename, "w") as ofile:
            ofile.write(str(self))

    def __str__(self) -> str:
        return "\n\n".join([str(s) for s in self._sections]) + "\n"

    def __repr__(self) -> str:
        return "%s(%s)" % (type(self).__name__, self.__str__())

    def __eq__(self, other) -> bool:
        if not isinstance(other, GDFile):
            return False
        return self._sections == other._sections

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)


class GDCommonFile(GDFile):
    """ Base class with common application logic for all Godot file types """

    def __init__(self, name: str, *sections: GDSection) -> None:
        super().__init__(
            GDSection(GDSectionHeader(name, load_steps=1, format=2)), *sections
        )
        self.load_steps = (
            1
            + len(self.get_sections("ext_resource"))
            + len(self.get_sections("sub_resource"))
        )

    @property
    def load_steps(self) -> int:
        return self._sections[0].header["load_steps"]

    @load_steps.setter
    def load_steps(self, steps: int):
        self._sections[0].header["load_steps"] = steps

    def add_section(self, new_section: GDSection) -> int:
        idx = super().add_section(new_section)
        if new_section.header.name in ["ext_resource", "sub_resource"]:
            self.load_steps += 1
        return idx

    def remove_at(self, index: int):
        section = self._sections.pop(index)
        if section.header.name in ["ext_resource", "sub_resource"]:
            self.load_steps -= 1
            self._recalculate_resource_ids(section.header.name)

    def _recalculate_resource_ids(self, section_name):
        """ Keep resource IDs in sync after removing one """
        id_map = {}
        # First we renumber all the resource IDs so there are no gaps
        for i, section in enumerate(self.get_sections(section_name)):
            id_map[section.id] = i + 1
            section.id = i + 1
        class_type = ExtResource if section_name == "ext_resource" else SubResource

        def replace(value):
            if isinstance(value, class_type):
                value.id = id_map.get(value.id)
            elif isinstance(value, list):
                for v in value:
                    replace(v)
            elif isinstance(value, dict):
                for v in value.values():
                    replace(v)
            elif isinstance(value, GDObject):
                for v in value.args:
                    replace(v)

        # Now we recursively traverse all nodes and update the resource IDs to stay in
        # sync with the renumbered resources
        for node in self.get_sections("node"):
            replace(node.header.attributes)
            replace(node.properties)
        for node in self.get_sections("resource"):
            replace(node.properties)


class GDScene(GDCommonFile):
    def __init__(self, *sections: GDSection) -> None:
        super().__init__("gd_scene", *sections)


class GDResource(GDCommonFile):
    def __init__(self, *sections: GDSection) -> None:
        super().__init__("gd_resource", *sections)
